import functools
from abc import ABC

from ipywidgets import Dropdown

from lhcsmapi.gui.BaseModule import BaseModule
from lhcsmapi.gui.pc.FgcBrowserAction import FgcBrowserAction
from lhcsmapi.metadata.SignalMetadata import SignalMetadata


class FgcPmCircuitNameDropdownBaseModule(BaseModule, ABC):

    def __init__(self, circuit_type):
        super().__init__()

        self.circuit_type = circuit_type

        # Widget
        circuit_names = get_circuit_names_or_prefixes_to_display(circuit_type)
        description = 'Circuit name prefix:' if circuit_type in ['600A', 'IPQ', '80-120A', '60A'] else 'Circuit name:'

        self.select_circuit_name_or_prefix_dropdown = Dropdown(options=circuit_names, description=description,
                                                               disabled=False)
        # On select value change, update output
        self.select_circuit_name_or_prefix_dropdown.observe(
            functools.partial(FgcPmCircuitNameDropdownBaseModule.handle_circuit_name_dropdown_change, self=self))

        self.widget = self.select_circuit_name_or_prefix_dropdown

    def handle_circuit_name_dropdown_change(d, self):
        self.mediator.notify(self, FgcBrowserAction.CIRCUIT_NAME_DROPDOWN_CHANGE)

    def get_circuit_name_or_family(self):
        return self.select_circuit_name_or_prefix_dropdown.value


def get_circuit_names_or_prefixes_to_display(circuit_type):
    """ Function returning a list of circuit names or types if the list of names is very long
    Depending on the circuit type the list is modified accordingly:
    - independent circuits: RB, IT, IPD - full list of names
    - coupled circuits: RQ, 600A_RCBXHV, 600A_RCDO - list of reduced names to avoid duplications,
    e.g., RQ.A12 instead of RQD.A12 and RQF.A12
    - multiple circuits: 600A - family types in order to avoid very long lists,
    e.g. RU instead of RU.L4

    :param circuit_type: circuit type such as RB, RQ, 600A, IT, IPD, IPQ
    :type circuit_type: str
    :return: list of circuit names or types
    """
    if circuit_type in ['RB', 'IT']:
        return SignalMetadata.get_circuit_names(circuit_type)
    elif circuit_type == 'IPD':
        return SignalMetadata.get_circuit_names(['IPD2', 'IPD2_B1B2'])
    elif circuit_type == '60A':
        r60A_names = SignalMetadata.get_circuit_names('60A')
        r60A_names_prefix = list(set([r60A_name.split('.')[0] for r60A_name in r60A_names]))
        return sorted(r60A_names_prefix)
    elif circuit_type == '80-120A':
        r80_120A_names = SignalMetadata.get_circuit_names('80-120A')
        r80_120A_names_prefix = list(set([SignalMetadata.parse_80_120A_circuit_name_to_family_name(r80_120A_name)
                                          for r80_120A_name in r80_120A_names]))
        return sorted(r80_120A_names_prefix)
    elif circuit_type == 'IPQ':
        ipq_names = SignalMetadata.get_circuit_names(['IPQ2', 'IPQ4', 'IPQ8'])
        ipqs_names_prefix = list(set([ipq_name.split('.')[0] for ipq_name in ipq_names]))
        return sorted(ipqs_names_prefix)
    elif circuit_type == 'RQ':
        circuit_names = SignalMetadata.get_circuit_names('RQ')
        circuit_names_rqd = list(filter(lambda name: 'RQD' in name, circuit_names))
        circuit_names_rqd_set = {circuit_name.replace('RQD', 'RQ') for circuit_name in circuit_names_rqd}
        return sorted(list(circuit_names_rqd_set))
    elif circuit_type == '600A':
        circuit_names = SignalMetadata.get_circuit_names('600A')
        circuit_names_without_rcbx = list(filter(lambda name: 'RCBX' not in name, circuit_names))
        circuit_names_without_rcbx_rcd = list(filter(lambda name: 'RCD' not in name, circuit_names_without_rcbx))
        circuit_names_without_rcbx_rcd_rco = list(
            filter(lambda name: 'RCO' not in name, circuit_names_without_rcbx_rcd))
        return sorted(list({circuit_name.split('.')[0] for circuit_name in circuit_names_without_rcbx_rcd_rco}))
    elif circuit_type == '600A_RCBXHV':
        circuit_names = SignalMetadata.get_circuit_names('600A')
        circuit_names_rcbxh = list(filter(lambda name: 'RCBXH' in name, circuit_names))
        circuit_names_rcbxh_set = {circuit_name.replace('RCBXH', 'RCBX') for circuit_name in circuit_names_rcbxh}
        return sorted(list(circuit_names_rcbxh_set))
    elif circuit_type == '600A_RCDO':
        circuit_names = SignalMetadata.get_circuit_names('600A')
        circuit_names_rcd = list(filter(lambda name: 'RCD' in name, circuit_names))
        circuit_names_rcd_set = {circuit_name.replace('RCD', 'RC') for circuit_name in circuit_names_rcd}
        return sorted(list(circuit_names_rcd_set))
    else:
        raise KeyError('Circuit type {} is not supported'.format(circuit_type))
