import functools
from typing import List, Tuple, Union

import pandas as pd
from IPython.display import display
from ipywidgets import Select, Label, HBox, Text, Layout, HTML, VBox

from lhcsmapi.Time import Time
from lhcsmapi.gui.BaseModule import BaseModule
from lhcsmapi.gui.pc.FgcBrowserAction import FgcBrowserAction
from lhcsmapi.metadata.MappingMetadata import MappingMetadata
from lhcsmapi.metadata.SignalMetadata import SignalMetadata
from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder
from lhcsmapi.pyedsl.dbquery.DbQueryBuilder import explode


class FgcPmEventSelectBaseModule(BaseModule):
    """ Base class for search of FGC PM events by respective notebook browsers

    """

    def __init__(self, circuit_type):
        super().__init__()

        self.circuit_type = circuit_type
        self.fgc_pm_events = []

        self.fgc_pm_event_sel = Select(options=[], rows=10, disabled=False, layout=Layout(width='50%'))

        # On select value change, update output
        self.fgc_pm_event_sel.observe(functools.partial(FgcPmEventSelectBaseModule.handle_event_selection_change,
                                                        self=self), names=['value'])

        first_row = HBox(children=[Label(value='FGC PM Events:'), self.fgc_pm_event_sel])

        # Fifth row - search for
        self.time_range = Text(placeholder='Time range for search of FGC', description='', value='5')
        self.time_range.observe(functools.partial(FgcPmEventSelectBaseModule.handle_time_range_change, self=self))
        self.time_range_label = Label(value='')
        second_row = HBox(children=[Label(value='Time range for search of FGC PM events around selected event [s]'),
                                    self.time_range, self.time_range_label])

        self.widget = VBox(children=[first_row, second_row])

    @staticmethod
    def handle_event_selection_change(b, self):
        if self.mediator is not None:
            self.mediator.notify(self, FgcBrowserAction.EVENT_SELECTION_CHANGE)

    @staticmethod
    def handle_time_range_change(b, self):
        if self.mediator is not None:
            self.mediator.notify(self, FgcBrowserAction.TIME_RANGE_CHANGE)

    def update_event_selection(self):
        if self.fgc_pm_events:
            # Fix an occasional issue with widget updates
            fgc_pm_event = self.get_fgc_pm_event()

            circuit_name = fgc_pm_event[0]
            timestamp = fgc_pm_event[1]
            pm_event_timestamp = self.create_pm_circuit_name_timestamp_text(fgc_pm_event)

            if self.circuit_type == '600A':
                print_ee_info_600A(circuit_name, self.circuit_type, timestamp)

            self.display_qps_circuit_schematic(circuit_name)

            print('Selected FGC PM event %s.' % pm_event_timestamp)

            # Displaying FGC PM timestamps from the same sector
            sector = MappingMetadata.get_sector_name_for_circuit_name(circuit_name)
            print('FGC PM events in circuits of sector %s:' % sector)

            self.time_range_label = 'Searching for other PM events in the sector, please wait...'
            time_range = float(self.time_range.value)
            source_timestamp_df = query_all_fgc_pm_events(timestamp, duration=[(time_range, 's'), (time_range, 's')])
            display_fgc_pm_events_from_the_same_sector(sector, source_timestamp_df, timestamp)
            self.time_range_label = 'Search for PM events is completed.'

    def get_fgc_pm_event(self):
        if self.fgc_pm_events:
            index = 0 if self.fgc_pm_event_sel.index is None else self.fgc_pm_event_sel.index
            return self.fgc_pm_events[index]
        else:
            return None

    def display_qps_circuit_schematic(self, circuit_name: str):
        pass

    def create_pm_circuit_name_timestamp_text(self, fgc_pm_event):
        if len(fgc_pm_event) == 2:
            return '{}: {}'.format(fgc_pm_event[0], Time.to_string(fgc_pm_event[1]))
        else:
            return '{}: {}; {}: {}' \
                .format(fgc_pm_event[0], Time.to_string(fgc_pm_event[1]),
                        fgc_pm_event[2], Time.to_string(fgc_pm_event[3]))

    def convert_fgc_pm_events_to_options(self, fgc_pm_entries: List[List[Union[str, int]]]) -> List[str]:
        """ Method converting a list of source and timestamp tuples into a list of strings to be displayed

        :param fgc_pm_entries: a list of FGC PM source and timestamp tuples
        :return: list of strings with FGC PM source and timestamp
        """

        options = []
        for fgc_pm_entry in fgc_pm_entries:
            if len(fgc_pm_entry) == 2:
                options.append('%s: %s' % (fgc_pm_entry[0], Time.to_string(fgc_pm_entry[1])))
            else:
                options.append('%s: %s; %s: %s' % (fgc_pm_entry[0], Time.to_string(fgc_pm_entry[1]),
                                                   fgc_pm_entry[2], Time.to_string(fgc_pm_entry[3])))
        return options

    def get_selected_circuit_name(self, fgc_pm_event):
        return fgc_pm_event[0]

    def get_selected_timestamp(self, fgc_pm_event):
        return fgc_pm_event[1]


def print_ee_info_600A(circuit_family: str, circuit_type: str, timestamp_query: int) -> None:
    circuit_name = list(filter(lambda col: circuit_family in col, SignalMetadata.get_circuit_names('600A')))[0]

    if 'EE' in SignalMetadata.get_system_types_per_circuit_name(circuit_type, circuit_name, timestamp_query):
        print('Circuit family %s has an EE system.' % circuit_family)
    else:
        print('Circuit family %s does not have an EE system.' % circuit_family)


def query_all_fgc_pm_events(timestamp_fgc: int, duration: List[Tuple[float, str]]) -> pd.DataFrame:
    return QueryBuilder().with_pm() \
        .with_duration(t_start=timestamp_fgc, duration=duration) \
        .with_query_parameters(system='FGC', className='*', source='*') \
        .event_query().df


def display_fgc_pm_events_from_the_same_sector(sector: str, source_timestamp_df: pd.DataFrame, timestamp_fgc: int
                                               ) -> None:

    if not source_timestamp_df.empty:
        circuit_name_to_sector_df = MappingMetadata.read_circuit_name_to_sector()
        circuit_name_to_sector_df = circuit_name_to_sector_df[circuit_name_to_sector_df['Sector name'] == sector]
        circuit_name_to_sector_df['PC name'] = get_pc_to_circuit_metadata(sector, timestamp_fgc)
        circuit_name_to_sector_df = explode(circuit_name_to_sector_df, 'PC name')

        circuit_name_to_sector_df = circuit_name_to_sector_df[circuit_name_to_sector_df['PC name'] \
            .isin(source_timestamp_df['source'])]

        circuit_name_to_sector_df = circuit_name_to_sector_df \
            .merge(source_timestamp_df, left_on=['PC name'], right_on=['source']) \
            .drop(columns=['source', 'Circuit type Layout'], axis=1)

        circuit_name_to_sector_df['dt [ms]'] = (circuit_name_to_sector_df['timestamp'] - timestamp_fgc) / 1e6
        circuit_name_to_sector_df['timestamp'] = circuit_name_to_sector_df['timestamp'] \
            .apply(lambda col: Time.to_string_short(col))

        circuit_name_to_sector_df.rename(columns={'Circuit type SIGMON': 'Circuit type'}, inplace=True)
        circuit_name_to_sector_df.sort_values(by='dt [ms]', inplace=True)
        circuit_name_to_sector_df.reset_index(drop=True, inplace=True)

        display(HTML(circuit_name_to_sector_df.to_html()))


@functools.lru_cache(maxsize=8)
def get_pc_to_circuit_metadata(selected_sector: str, timestamp_query: int) -> List[List[str]]:
    """ Method returning power converter names belonging to a sector

    :param selected_sector: name of a selected sector
    :return: List of power converter names belonging to a sector
    """
    pc_to_circuit_metadata = []
    circuit_name_to_sector_df = MappingMetadata.read_circuit_name_to_sector()
    mask = circuit_name_to_sector_df['Sector name'] == selected_sector
    for index, row in circuit_name_to_sector_df[mask].iterrows():
        circuit_metadata = SignalMetadata.get_circuit_name_metadata(row['Circuit type SIGMON'], row['Circuit name'],
                                                                    timestamp_query=timestamp_query)
        pc_name = circuit_metadata['PC']['%PC%']
        if isinstance(pc_name, list):
            pc_to_circuit_metadata.append(pc_name)
        else:
            pc_to_circuit_metadata.append([pc_name])

    return pc_to_circuit_metadata
