from typing import Tuple, List

from lhcsmapi.gui.pc.fgc_pm_find_button.FgcPmFindButtonBaseModule import FgcPmFindButtonBaseModule
from lhcsmapi.metadata.SignalMetadata import SignalMetadata
from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder


class IpqFgcPmFindButtonBaseModule(FgcPmFindButtonBaseModule):
    """ A class with methods to search for IPQ FGC PM events to fill-in a search field in the FGC PM search module.

    """

    def find_fgc_pm_events_with_qps(self, start_time, end_time, circuit_name: str) -> List[Tuple[str, int]]:
        """ Method finding FGC PM entries with QPS for a given time range and circuit name

        :param start_time: start time for the query
        :param end_time: end time for the query
        :param circuit_name: circuit name
        :return: a list of FGC source and timestamp tuples
        """
        self.lbl_search.value = 'Querying PM for FGC events'
        self.progress_search.layout.visibility = 'visible'
        self.progress_search.max = 1
        self.progress_search.description = ''

        output = query_pm_convert_pc_to_circuit_name(start_time, end_time, circuit_name)

        self.lbl_search.value = 'Query completed'
        self.progress_search.layout.visibility = 'visible'
        self.progress_search.value = 1
        self.progress_search.description = ''

        return output


def query_pm_convert_pc_to_circuit_name(start_time, end_time, circuit_name: str) -> List[Tuple[str, int]]:
    """ Method querying FGC PM source and timestamp tuple and taking only those corresponding to the given circuits

    :param start_time: start time for a query
    :param end_time: end time for a query
    :param circuit_name: circuit name
    :return: a list of PM source and timestamp tuples corresponding to circuit names
    """
    # It is assumed that all IPQ circuits have the same FGC PC className
    metadata = SignalMetadata.get_circuit_signal_database_metadata('IPQ2', 'RQ5.R5', 'PC', 'PM',
                                                                   timestamp_query=start_time)
    source_timestamp_df = QueryBuilder().with_pm() \
        .with_duration(t_start=start_time, t_end=end_time) \
        .with_query_parameters(system='FGC', className=metadata['className'], source='*') \
        .event_query().drop_duplicates('timestamp').df

    if not source_timestamp_df.empty:
        pc_names = SignalMetadata.get_pc_names_for_circuit_type(['IPQ2', 'IPQ4', 'IPQ8'], start_time)
        source_timestamp_df = source_timestamp_df[source_timestamp_df['source'].isin(pc_names)]
        source_timestamp_df = source_timestamp_df[source_timestamp_df['source'].str.contains(circuit_name)]

    # Convert PC names to circuit names
    output = []
    for index, row in source_timestamp_df.iterrows():
        pc_name = row['source']
        pc_name_split = pc_name.split('.')
        circuit_name = '%s.%s' % (pc_name_split[-2], pc_name_split[-1][0:2])
        output.append((circuit_name, row['timestamp']))

    return output
