from typing import List, Tuple, Union, Any

from lhcsmapi.Time import Time
from lhcsmapi.metadata.SignalMetadata import SignalMetadata
from lhcsmapi.pyedsl.dbsignal.post_mortem.PmDbRequest import PmDbRequest
from lhcsmapi.gui.pc.fgc_pm_find_button.FgcPmFindButtonBaseModule import FgcPmFindButtonBaseModule


class RqFgcPmFindButtonBaseModule(FgcPmFindButtonBaseModule):
    """ A class with methods to search for RQ FGC PM events to fill-in a search field in the FGC PM search module.

    """

    def find_fgc_pm_events_with_qps(self, start_time, end_time, circuit_name: str
                                    ) -> List[List[Union[Union[str, int], Any]]]:
        """ Method finding FGC PM entries with QPS for a given time range and circuit name

        :param start_time: start time for the query
        :param end_time: end time for the query
        :param circuit_name: circuit name
        :return: a list of FGC source and timestamp tuples
        """
        # convert RQ.AXX into [RQD.AXX, RQF.AXX]
        circuit_names = [circuit_name.replace('RQ', rq_circuit_name) for rq_circuit_name in ['RQD', 'RQF']]

        source_timestamp_fgc = self._find_fgc_pm_entries(start_time, end_time, 'RQ', circuit_names)
        return self._keep_fgc_rqd_with_rqf('RQ', circuit_names, source_timestamp_fgc)

    def _find_fgc_pm_entries(self, start_time, end_time, circuit_type: str, circuit_names: List[str]
                             ) -> List[Tuple[str, int]]:
        """ Method finding FGC PM entries for a given start time and end time

        :param start_time: query start time
        :param end_time: query end time
        :param circuit_type: circuit type
        :param circuit_names: list of circuit names
        :return: a list of source and timestamp tuples
        """

        metadata_fgc_rqd = SignalMetadata.get_circuit_signal_database_metadata(circuit_type, circuit_names[0], 'PC',
                                                                               'PM', timestamp_query=start_time)

        quotient, remainder = Time.modulo_day_in_unix_timestamp(start_time, end_time)
        event_duration_nsec = int(24 * 60 * 60 * 1e9)
        source_timestamp_fgc = []

        self.progress_search.layout.visibility = 'visible'
        self.progress_search.max = quotient
        self.progress_search.description = f'Days: 0/{quotient}'

        for day in range(quotient):
            start_time_temp = Time.to_unix_timestamp(start_time) + int(day * event_duration_nsec)

            # Update GUI
            self.lbl_search.value = 'Querying PM for FGC entries on {}'.format(Time.to_string(start_time_temp))

            source_timestamp_fgc += PmDbRequest.find_events(metadata_fgc_rqd['source'], metadata_fgc_rqd['system'],
                                                            metadata_fgc_rqd['className'], t_start=start_time_temp,
                                                            duration=[(event_duration_nsec, 'ns')])
            self.progress_search.description = f'Days: {day + 1}/{quotient}'
            self.progress_search.value = day + 1

        if remainder > 0:
            start_time_temp = Time.to_unix_timestamp(end_time) - remainder
            self.lbl_search.value = 'Querying PM for FGC entries on {}'.format(Time.to_string(start_time_temp))
            source_timestamp_fgc += PmDbRequest.find_events(metadata_fgc_rqd['source'], metadata_fgc_rqd['system'],
                                                            metadata_fgc_rqd['className'], t_start=start_time_temp,
                                                            duration=[(remainder, 'ns')])

        return source_timestamp_fgc

    def _keep_fgc_rqd_with_rqf(self, circuit_type: str, circuit_names: List[str],
                                       source_timestamp_fgc: List[Tuple[str, int]]
                                       ) -> List[Tuple[Union[Union[str, int], Any], ...]]:
        """ Method keeping FGC PM events with RQF

        :param circuit_type: circuit type
        :param circuit_names: List of RQ circuit names (RQD.AXY, RQF.AXY)
        :param source_timestamp_fgc: a list of tuples with source and timestamp
        :return: a list of tuples with source and timestamp that has an QPS PM event around an FGC as well as a
        corresponding RQF FGC timestamp
        """
        # Check if an RQD FGC timestamp has
        # # an RQF FGC timestamp
        source_timestamp_fgc_with_rqf = []

        for index, elem in enumerate(source_timestamp_fgc):
            source_fgc, timestamp_fgc = elem
            metadata_fgc_rqf = SignalMetadata.get_circuit_signal_database_metadata(circuit_type, circuit_names[1], 'PC',
                                                                                   'PM', timestamp_query=timestamp_fgc)
            # Update GUI
            self.lbl_search.value = 'Checking if there are QPS PM around an FGC timestamp {}'.format(
                Time.to_string(elem[1]))
            self.progress_search.description = f'FGC: {index + 1}/{len(source_timestamp_fgc)}'
            self.progress_search.value = index + 1

            # Check if there is a corresponding RQF timestamp
            source_timestamp_fgc_rqf = PmDbRequest.find_events(metadata_fgc_rqf['source'], metadata_fgc_rqf['system'],
                                                               metadata_fgc_rqf['className'], t_start=timestamp_fgc,
                                                               duration=[(1, 's'), (1, 's')])
            if source_timestamp_fgc_rqf:
                source_timestamp_fgc_with_rqf.append(
                    tuple([circuit_names[0], timestamp_fgc, circuit_names[1], source_timestamp_fgc_rqf[0][1]]))

        self.lbl_search.value = 'Query completed.'

        return source_timestamp_fgc_with_rqf
