from datetime import datetime

import numpy as np
import pandas as pd

from lhcsmapi.Time import Time
from lhcsmapi.pyedsl.KeyTypeValidation import KeyTypeValidation
from lhcsmapi.pyedsl.dbsignal.SignalDbQuery import SignalDbQuery
from lhcsmapi.pyedsl.dbsignal.nxcals.NxcalsDevicePropertySignalDbQuery import NxcalsDevicePropertySignalDbQuery
from lhcsmapi.pyedsl.dbsignal.nxcals.NxcalsVariableSignalDbQuery import NxcalsVariableSignalDbQuery
from lhcsmapi.pyedsl.dbsignal.post_mortem.PmSignalDbQuery import PmSignalDbQuery


class SignalDbQueryFactory:
    """ Class builds objects for querying logging databases with signals

    """

    @classmethod
    def build(cls, api, **kwargs) -> SignalDbQuery:
        """ Method building a SignalDbQuery object for either PM or NXCALS. For PM, PmSignalDbQuery is returned.
        For NXCALS there are two types available:
        - for a variable query, NxcalsVariableSignalDbQuery is returned
        - for device and property query, NxcalsDevicePropertySignalDbQuery is returned

        :param api: DB API (either PM or NXCALS)
        :param kwargs: a dictionary with parameters of a query
        :return: a SignalDbQuery object matching input parameters
        """
        if api.upper() == "PM":
            return cls.build_pm(**kwargs)
        elif api.upper() == "NXCALS":
            if "nxcals_device" in kwargs and "nxcals_property" in kwargs:
                return cls.build_nxcals_device_property(**kwargs)
            elif not "nxcals_device" in kwargs and not "nxcals_property" in kwargs:
                return cls.build_nxcals_variable(**kwargs)
            else:
                raise TypeError('{} --> kwargs definition does not match any known NXCALS type'.format(kwargs))
        else:
            raise TypeError('{} key does not match any known APIs: PM, NXCALS'.format(api.upper()))

    @classmethod
    def build_pm(cls, **kwargs) -> PmSignalDbQuery:
        """Method building a PmSignalDbQuery from an input (keyword: argument) dictionary.
        It validates input kwargs and implements static builder design pattern to build a new PmSignalDbQuery object.
        If arguments are not provided, default arguments are used

        :param kwargs: dictionary of (keyword: argument) pairs which parametrize a REST PM API query
        :type kwargs: dictionary
        :return: initialized PmSignalDbQuery object

        - Example::

            >>> from lhcsmapi.pyedsl.dbsignal.post_mortem.PmSignalDbQuery import PmSignalDbQuery
            >>> system = "QPS"
            >>> class_name = "DQAMCNMB_PMHSU"
            >>> source = "B20L5"
            >>> signal = "I_HDS_4"
            >>> eventTime = 1426220469491000000
            >>> str(PmSignalDbQuery(system=system, className=class_name, source=source,
            ...              signal=signal, evenTime=event_time))
            'Db: PM, signal: I_HDS_4, event_time = 1426220469491000000 (2015-03-13 05:21:09.491000+01:00), system=QPS, source=B20L5, class_name=DQAMCNMB_PMHSU'
        """

        system = KeyTypeValidation.validate_key_in_kwargs('system', str, **kwargs) or None
        class_name = KeyTypeValidation.validate_key_in_kwargs('className', str, **kwargs) or None
        source = KeyTypeValidation.validate_key_in_kwargs('source', str, **kwargs) or None
        signal = KeyTypeValidation.validate_key_in_kwargs('signal', str, **kwargs) or None
        event_time = KeyTypeValidation.validate_key_in_kwargs('eventTime', int, pd.Timestamp, np.int64,
                                                              **kwargs) or None
        pm_rest_api_path = KeyTypeValidation.validate_key_in_kwargs('pm_rest_api_path', str, **kwargs) or None
        show_warning = KeyTypeValidation.validate_key_in_kwargs('show_warning', bool, **kwargs) or True
        allow_duplicate_index = KeyTypeValidation.validate_key_in_kwargs('allow_duplicate_index', bool,
                                                                         **kwargs) or False

        return PmSignalDbQuery(system, class_name, source, signal, event_time,
                               pm_rest_api_path, show_warning, allow_duplicate_index)

    @classmethod
    def build_nxcals_device_property(cls, **kwargs) -> NxcalsDevicePropertySignalDbQuery:
        """Method building a NxcalsDevicePropertySignalDbQuery from an input (keyword: argument) dictionary.
        It validates input kwargs and implements static builder design pattern to build a new
        NxcalsDevicePropertySignalDbQuery object.
        If arguments are not provided, default arguments are used.

        :param kwargs: dictionary of (keyword: argument) pairs which parametrize an NXCALS API query
        :type kwargs: dictionary
        :return: initialized NxcalsDevicePropertySignalDbQuery object
        """

        t_start = KeyTypeValidation.validate_key_in_kwargs('t_start', int, pd.Timestamp, str, float, datetime, tuple,
                                                           np.datetime64, **kwargs) or 0
        t_end = KeyTypeValidation.validate_key_in_kwargs('t_end', int, pd.Timestamp, str, float, datetime, tuple,
                                                         np.datetime64, **kwargs) or None
        duration = KeyTypeValidation.validate_key_in_kwargs('duration', tuple, list, **kwargs) or None
        tz = KeyTypeValidation.validate_key_in_kwargs('tz', str, **kwargs) or None

        nxcals_system = KeyTypeValidation.validate_key_in_kwargs('nxcals_system', str, **kwargs) or None
        nxcals_devices = KeyTypeValidation.validate_key_in_kwargs('nxcals_device', str, list, **kwargs) or []
        nxcals_property = KeyTypeValidation.validate_key_in_kwargs('nxcals_property', str, **kwargs) or None

        # we need to take the first element from the list as signal
        signals = KeyTypeValidation.validate_key_in_kwargs('signals', list, **kwargs) or []

        # calculate start and end time based of parsed arguments
        t_start, t_end = Time.get_query_period_in_unix_time(t_start, t_end, duration, tz)

        # convert start and end to unix time in seconds with us precision for NXCALS api
        t_start_nxcals = pd.Timestamp(t_start).to_datetime64()
        t_end_nxcals = pd.Timestamp(t_end).to_datetime64()

        spark = kwargs['spark']

        return NxcalsDevicePropertySignalDbQuery(t_start=t_start_nxcals,
                                                 t_end=t_end_nxcals,
                                                 signal=signals[0],
                                                 nxcals_system=nxcals_system,
                                                 nxcals_devices=nxcals_devices,
                                                 nxcals_property=nxcals_property,
                                                 spark=spark)

    @classmethod
    def build_nxcals_variable(cls, **kwargs) -> NxcalsVariableSignalDbQuery:
        """Method building a NxcalsVariableSignalDbQuery from an input (keyword: argument) dictionary.
        It validates input kwargs and implements static builder design pattern to build a new
        NxcalsVariableSignalDbQuery object.
        If arguments are not provided, default arguments are used.

        :param kwargs: dictionary of (keyword: argument) pairs which parametrize an NXCALS API query
        :type kwargs: dictionary
        :return: initialized NxcalsVariableSignalDbQuery object
        """

        t_start = KeyTypeValidation.validate_key_in_kwargs('t_start', int, pd.Timestamp, str, float, datetime, tuple,
                                                           np.datetime64, **kwargs) or 0
        t_end = KeyTypeValidation.validate_key_in_kwargs('t_end', int, pd.Timestamp, str, float, datetime, tuple,
                                                         np.datetime64, **kwargs) or None
        duration = KeyTypeValidation.validate_key_in_kwargs('duration', tuple, list, **kwargs)
        tz = KeyTypeValidation.validate_key_in_kwargs('tz', str, **kwargs)

        nxcals_system = KeyTypeValidation.validate_key_in_kwargs('nxcals_system', str, **kwargs) or None
        signals = KeyTypeValidation.validate_key_in_kwargs('signals', list, **kwargs) or None

        # calculate start and end time based of parsed arguments
        t_start, t_end = Time.get_query_period_in_unix_time(t_start, t_end, duration, tz)

        # convert start and end to unix time in seconds with us precision for CALS api
        t_start_nxcals = pd.Timestamp(t_start).to_datetime64()
        t_end_nxcals = pd.Timestamp(t_end).to_datetime64()

        spark = kwargs['spark']

        return NxcalsVariableSignalDbQuery(t_start=t_start_nxcals,
                                           t_end=t_end_nxcals,
                                           nxcals_system=nxcals_system,
                                           signals=signals,
                                           spark=spark)
