from typing import List
import warnings

import pandas as pd

from lhcsmapi.pyedsl.dbsignal.SignalDbQuery import SignalDbQuery

try:
    from cern.nxcals.pyquery.builders import DevicePropertyQuery
except:
    pass


class NxcalsDevicePropertySignalDbQuery(SignalDbQuery):
    """NxcalsDevicePropertySignalDbQuery provides a signal query to NXCALS with DevicePropertyQuery builder.
    """

    def __init__(self,
                 t_start=0,
                 t_end=0,
                 signal=None,
                 nxcals_system=None,
                 nxcals_devices=[],
                 nxcals_property=None,
                 spark=None):
        """ Constructor of an NxcalsDevicePropertySignalDbQuery object

        :param t_start: query start time
        :param t_end: query end time
        :param signal: signal name
        :param nxcals_system: NXCALS system name
        :param nxcals_devices: list of NXCALS device names
        :param nxcals_property: NXCALS property name
        :param spark: spark connector
        """
        super().__init__('NXCALS')
        self.t_start = t_start
        self.t_end = t_end
        self.signal = signal
        self.nxcals_system = nxcals_system
        self.nxcals_devices = nxcals_devices
        self.nxcals_property= nxcals_property
        self.spark = spark

        if self.spark is None:
            raise Exception('You have to provide an instance of spark connector via the spark keyword')

    def read(self) -> List[pd.DataFrame]:
        """ Method querying a signal from NXCALS between start and end time for a given system, device, and property.
        The method uses DevicePropertyQuery from NXCALS API.
        In case of an exception, an appropriate warning is displayed and a list of empty DataFrames is returned.

        :return: a list of DataFrame with a requested signals (one per device)
        """

        try:
            # Initialize builder
            device_property_builder = DevicePropertyQuery.builder(self.spark) \
                .system(self.nxcals_system) \
                .startTime(self.t_start) \
                .endTime(self.t_end) \

            # Populate builder
            for nxcals_device in self.nxcals_devices:
                device_property_builder = device_property_builder.entity() \
                    .device(nxcals_device).property(self.nxcals_property)

            # Execute query
            multisignal_df = device_property_builder.buildDataset().select(
                ['acqStamp', self.signal, 'device']).dropna().toPandas()
            multisignal_df.set_index(multisignal_df['acqStamp'], inplace=True)
            multisignal_df.drop(columns='acqStamp', inplace=True)

            multisignal_dct = dict(tuple(multisignal_df.groupby('device')))
            signal_dfs = []
            for nxcals_device in self.nxcals_devices:
                signal_df = multisignal_dct.get(nxcals_device, pd.DataFrame(columns=[self.signal]))
                signal_df.sort_index(inplace=True)
                signal_df.drop(columns='device', inplace=True)
                signal_dfs.append(signal_df)

            return signal_dfs

        except Exception as e:
            warning = """NXCALS query failed with the parameters, nxcals_system: {0}, t_start: {1}, t_end: {2}, 
                        nxcals_devices: {3}, nxcals_property: {4}, signal: {5}\n
                        List of empty signals returned.
                        ___________________________________________________________________________________
                        You can reproduce the error using the raw NXCALS API:\n
                        from cern.nxcals.pyquery.builders import *
                        
                        device_property_builder = DevicePropertyQuery.builder(spark) \\
                            .system({0}) \\
                            .startTime({1}) \\
                            .endTime({2}) \\

                        # Populate builder
                        for nxcals_device in {3}:
                            device_property_builder = device_property_builder.entity() \\
                                .device(nxcals_device).property({4})
            
                        # Execute query
                        multisignal_df = device_property_builder.buildDataset().select(
                            ['acqStamp', {5}, 'device']).dropna().toPandas()
                        multisignal_df.set_index(multisignal_df['acqStamp'], inplace=True)
                        multisignal_df.drop(columns='acqStamp', inplace=True)
                        
                        # Extract signals in the order of input devices
                        multisignal_dct = dict(tuple(multisignal_df.groupby('device')))
                        signal_dfs = []
                        for nxcals_device in {3}:
                            signal_df = multisignal_dct.get(nxcals_device, pd.DataFrame(columns=[{5}]))
                            signal_df.sort_index(inplace=True)
                            signal_df.drop(columns='device')
                            signal_dfs.append(signal_df)
                        """.format(self.nxcals_system, self.t_start, self.t_end, self.nxcals_devices,
                                   self.nxcals_property, self.signal)

            warnings.warn(str(e))
            warnings.warn(warning)
            return [pd.DataFrame(columns=[self.signal]) for _ in range(len(self.nxcals_devices))]
