import warnings
from typing import List

import pandas as pd

from lhcsmapi.pyedsl.dbsignal.SignalDbQuery import SignalDbQuery

try:
    from cern.nxcals.pyquery.builders import VariableQuery
except:
    pass


class NxcalsVariableSignalDbQuery(SignalDbQuery):
    """NxcalsVariableSignalDbQuery provides a signal query to NXCALS with VariableQuery builder.
    """


    def __init__(self,
                 t_start=0,
                 t_end=0,
                 nxcals_system=None,
                 signals=[],
                 spark=None):
        """ Constructor of an NxcalsVariableSignalDbQuery object

        :param t_start: query start time
        :param t_end: query end time
        :param signals: list of signal names
        :param nxcals_system: NXCALS system name
        :param spark: spark connector
        """
        super().__init__('NXCALS')
        self.t_start = t_start
        self.t_end = t_end
        self.nxcals_system = nxcals_system
        self.signals = signals
        self.spark = spark

        if self.spark is None:
            raise Exception('You have to provide an instance of spark connector via the spark keyword')

    def read(self) -> List[pd.DataFrame]:
        """ Method querying a signal from NXCALS between start and end time for a given system.
        The method uses VariableQuery from NXCALS API.
        In case of an exception, an appropriate warning is displayed and a list of empty DataFrames is returned.

        :return: a list of DataFrame with a requested signals (one per device)
        """
        try:
            variable_builder = VariableQuery.builder(self.spark) \
                .system(self.nxcals_system) \
                .startTime(self.t_start).endTime(self.t_end)

            for signal in self.signals:
                variable_builder = variable_builder.variable(signal)

            multisignal_df = variable_builder.buildDataset().select(
                ['nxcals_timestamp', 'nxcals_value', 'nxcals_variable_name']).dropna().toPandas()

            multisignal_df.set_index(multisignal_df['nxcals_timestamp'], inplace=True)
            multisignal_df.drop(columns='nxcals_timestamp', inplace=True)
            multisignal_df.dropna(inplace=True)
            multisignal_df.sort_index(inplace=True)

            multisignal_dct = dict(tuple(multisignal_df.groupby('nxcals_variable_name')))
            signal_dfs = []
            for signal in self.signals:
                signal_df = multisignal_dct.get(signal, pd.DataFrame(columns=[signal]))
                signal_df.sort_index(inplace=True)
                signal_df.drop(columns='nxcals_variable_name', inplace=True)
                signal_df.rename(columns={'nxcals_value': signal}, inplace=True)
                signal_dfs.append(signal_df)

            return signal_dfs

        except Exception as e:
            warning = """NXCALS query failed with the parameters, nxcals_system: {0}, t_start: {1}, t_end: {2}, 
                        signals: {3}\n
                        List of empty signals returned.
                        ___________________________________________________________________________________
                        You can reproduce the error using the raw NXCALS API:\n
                        from cern.nxcals.pyquery.builders import *
            
                        variable_builder = VariableQuery.builder(spark) \\
                            .system({0}) \\
                            .startTime({1}).endTime({2})

                        for signal in {3}:
                            variable_builder = variable_builder.variable(signal)
            
                        multisignal_df = variable_builder.buildDataset().select(
                            ['nxcals_timestamp', 'nxcals_value', 'nxcals_variable_name']).dropna().toPandas()
            
                        multisignal_df.set_index(multisignal_df['nxcals_timestamp'], inplace=True)
                        multisignal_df.drop(columns='nxcals_timestamp', inplace=True)
                        multisignal_df.dropna(inplace=True)
                        multisignal_df.sort_index(inplace=True)
            
                        multisignal_dct = dict(tuple(multisignal_df.groupby('nxcals_variable_name')))
                        signal_dfs = []
                        for signal in {3}:
                            signal_df = multisignal_dct.get(signal, pd.DataFrame(columns=[signal]))
                            signal_df.sort_index(inplace=True)
                            signal_df.drop(columns='nxcals_variable_name', inplace=True)
                            signal_df.rename(columns=dict('nxcals_value': signal), inplace=True)
                            signal_dfs.append(signal_df)
                """.format(self.nxcals_system, self.t_start, self.t_end, self.signals)

            warnings.warn(str(e))
            warnings.warn(warning)
            return [pd.DataFrame(columns=[signal]) for signal in range(len(self.signals))]
