from typing import List

import pandas as pd

from lhcsmapi.Time import Time
from lhcsmapi.pyedsl.dbsignal.SignalDbQuery import SignalDbQuery
from lhcsmapi.pyedsl.dbsignal.post_mortem.PmDbRequest import PmDbRequest


class PmSignalDbQuery(SignalDbQuery):
    """PmSignalDbQuery class parses a JSON signal obtained with the Post Mortem REST API.
    """

    def __init__(self, system=None, className=None, source=None, signal=None, eventTime=None,
                 pm_rest_api_path=None, show_warning=True, allow_duplicate_index=False):
        super().__init__('PM')
        self.system = system
        self.class_name = className
        self.source = source
        self.signal = signal
        self.event_time = eventTime
        self.pm_rest_api_path = pm_rest_api_path
        self.show_warning = show_warning
        self.allow_duplicate_index = allow_duplicate_index
        self.time = []
        self.value = []

    def __str__(self):
        return "Db: PM, signal: {}, eventTime = {} ({}), system={}, source={}, className={}". \
            format(self.signal, self.event_time, Time.to_string(self.event_time, 'ns'),
                   self.system, self.source, self.class_name)

    def read(self) -> List[pd.DataFrame]:
        """Method returning a list of pandas Dataframe containing a signal with an index formatted to ns.

        :param kwargs: dictionary of (keyword: argument) pairs which parametrize an db API query (optional)
        :type kwargs: dictionary
        :return: initialized PmDbSignal object

        - Example::

            >>> from lhcsmapi.pyedsl.dbsignal.post_mortem.PmSignalDbQuery import PmSignalDbQuery
            >>> system = "QPS"
            >>> className = "DQAMCNMB_PMHSU"
            >>> source = "B20L5"
            >>> signal = "I_HDS_4"
            >>> eventTime = 1426220469491000000
            >>> pms = PmSignalDbQuery(system=system, className=className, source=source,
            ...                  signal=signal, eventTime=eventTime) # doctest: +SKIP
            >>> I_HDS_4 = pms.read() # doctest: +SKIP
            >>> print(I_HDS_4.head()) # doctest: +SKIP
                                  I_HDS_4
            1426220469491020832  0.089648
            1426220469491026040  0.091555
            1426220469491031248  0.091555
            1426220469491036456  0.093463
            1426220469491041664  0.091555
        """
        try:
            self._parse_signal()
            df = self._get_signal_df(self.signal)
            # remove 0-valued indices
            df = df.loc[df.index != 0]
            # remove duplicated indices
            if not self.allow_duplicate_index:
                df = df.loc[~df.index.duplicated(keep=False)]
                # remove non-monotonic indices
                df['index'] = df.index
                df['diff_index'] = df['index'].diff()
                df.fillna(1, inplace=True)
                df = df[df['diff_index'] > 0]
                df.drop(columns=['index', 'diff_index'], inplace=True)

            return [df]
        except:
            return [pd.DataFrame()]

    def _parse_signal(self):
        """Method looking for parameter and time in order to return a Dataframe with the signal.
        """
        response = PmDbRequest.get_response("pmdata/signal", False, self.show_warning, self.pm_rest_api_path,
                                            system=self.system, className=self.class_name, source=self.source,
                                            timestampInNanos=self.event_time, signal=self.signal)

        content = response['content'][0]

        # find SIGNAL_NAMES
        name_of_value = PmSignalDbQuery._get_names_of_value(self, content)

        # find time and value
        PmSignalDbQuery._get_time_value(self, name_of_value, content)

    def _get_signal_df(self, *columns):
        """Method returning a pandas DataFrame based on internal time and value properties.
        """
        return pd.DataFrame(self.value, self.time, columns=[col for col in columns])

    def _get_names_of_value(self, content):
        """Method returning a list of SIGNAL_NAMES contained in content.
        """

        size = len(content['namesAndValues'])
        name_of_value = ["" for x in range(0, size)]
        name_of_value[:] = [content['namesAndValues'][n]['name'] for n in range(0, size)]

        return name_of_value

    def _get_time_value(self, name_of_value, content):
        """Method setting time (in ns) and value from the content received with a PM signal query.
        """
        start_time, interv_time = None, None
        self.event_time = content['rawDataEntry']['pmEventStamp']
        for s in range(0, len(name_of_value)):
            if "START_TIME_NSEC" in name_of_value[s]:
                start_time = content['namesAndValues'][s]['value']
            elif "INTERVAL_NSEC" in name_of_value[s]:
                interv_time = content['namesAndValues'][s]['value']
            elif any(keyword in name_of_value[s] for keyword in ["AQNTIME", "TIMESTAMP"]):
                self.time = content['namesAndValues'][s]['value']
            elif self.signal in name_of_value[s]:  # search for SignalName
                self.value = content['namesAndValues'][s]['value']
            else:
                raise ValueError("Keyword {} not supported for this type of query.".format(name_of_value[s]))

        if not self.time:  # Time empty, has to be generated from t_start and interv_time
            self.time = [t for t in range(start_time, start_time + interv_time * len(self.value), interv_time)]
