import unittest
import warnings

import pandas as pd

from lhcsmapi.reference.Reference import Reference


class TestReference(unittest.TestCase):

    def test_get_quench_heater_reference_discharges_RB(self):
        # arrange
        cell = "A8R1"
        circuit_type = 'RB'

        # act
        ref_discharge_df = Reference.read_quench_heater_reference_discharges(circuit_type)
        timestamp_act = ref_discharge_df[ref_discharge_df["Cell"] == cell]["Timestamp"].values[0]

        # assert
        timestamp_exp = 1417014253613000000
        self.assertEqual(timestamp_exp, timestamp_act)

    def test_get_quench_heater_reference_discharges_RQ(self):
        # arrange
        cell = "16L2"
        circuit_type = 'RQ'

        # act
        ref_discharge_df = Reference.read_quench_heater_reference_discharges(circuit_type)
        timestamp_act = ref_discharge_df[ref_discharge_df["Cell"] == cell]["Timestamp"].values[0]

        # assert
        timestamp_exp = 1417079205409000000
        self.assertEqual(timestamp_exp, timestamp_act)

    def test_get_quench_heater_reference_discharge_RQ(self):
        # arrange
        cell = "16L2"
        circuit_type = 'RQ'

        # act
        timestamp_act = Reference.get_quench_heater_reference_discharge(circuit_type,
                                                                        cell,
                                                                        1417079205509000000)

        # assert
        timestamp_exp = 1417079205409000000
        self.assertEqual(timestamp_exp, timestamp_act)

    def test_get_quench_heater_reference_discharge_error(self):
        # arrange
        cell = ""
        circuit_type = '600A'

        # act
        with self.assertRaises(ValueError) as context:
            Reference.get_quench_heater_reference_discharge(circuit_type,
                                                            cell,
                                                            1)

        # assert
        self.assertEqual('Circuit type 600A not supported!', str(context.exception))

    def test_read_energy_extraction_reference_features_rb(self):
        # arrange
        circuit_type = 'RB'

        # act
        features_df_act = Reference.read_energy_extraction_reference_features(circuit_type)

        # assert
        features_df_ref = pd.DataFrame({'min': {'R_odd': 0.0675, 'R_even': 0.0675, 't_delay_ee_odd': 0.05,
                                                't_delay_ee_even': 0.55, 'tau_u_dump_res_odd': 110.0,
                                                'tau_u_dump_res_even': 110.0},
                                        'max': {'R_odd': 0.0825, 'R_even': 0.0825, 't_delay_ee_odd': 0.15,
                                                't_delay_ee_even': 0.65, 'tau_u_dump_res_odd': 130.0,
                                                'tau_u_dump_res_even': 130.0}})
        pd.testing.assert_frame_equal(features_df_ref, features_df_act)

    def test_read_energy_extraction_reference_features_error(self):
        # arrange
        circuit_type = '600A'

        # act
        with self.assertRaises(KeyError) as context:
            Reference.read_energy_extraction_reference_features(circuit_type)

        # assert
        self.assertEqual('\'Circuit type 600A not supported!\'', str(context.exception))

    def test_read_power_converter_reference_features_rb(self):
        # arrange
        circuit_type = 'RB'

        # act
        features_df_act = Reference.read_power_converter_reference_features(circuit_type)

        # assert
        features_df_ref = pd.DataFrame({'min': {'tau_i_meas': 90}, 'max': {'tau_i_meas': 110}})
        pd.testing.assert_frame_equal(features_df_ref, features_df_act)

    def test_read_power_converter_reference_features_error(self):
        # arrange
        circuit_type = 'IT'

        # act
        with self.assertRaises(ValueError) as context:
            Reference.read_power_converter_reference_features(circuit_type)

        # assert
        self.assertEqual('Reference features for IT circuit type are not defined.', str(context.exception))

    def test_get_power_converter_reference_fpa_rb(self):
        # arrange
        circuit_type = 'RB'
        circuit_name = 'RB.A12'

        # act
        fgc_pm_act = Reference.get_power_converter_reference_fpa(circuit_type, circuit_name, col='fgcPm')

        # assert
        fgc_pm_ref = '2018-03-17 16:00:28.680'
        self.assertEqual(fgc_pm_ref, fgc_pm_act)

    def test_get_power_converter_reference_fpa_rq(self):
        # arrange
        circuit_type = 'RQ'
        circuit_name = 'RQD.A12'

        # act
        fgc_pm_act = Reference.get_power_converter_reference_fpa(circuit_type, circuit_name, col='fgcPm')

        # assert
        fgc_pm_ref = '2018-03-18 11:58:33.600'
        self.assertEqual(fgc_pm_ref, fgc_pm_act)

    def test_get_power_converter_reference_fpa_it_error(self):
        # arrange
        circuit_type = 'IT'
        circuit_name = 'RQX.L1'

        # act
        with self.assertRaises(ValueError) as context:
            Reference.get_power_converter_reference_fpa(circuit_type, circuit_name, col='fgcPm')

        # assert
        self.assertEqual('Circuit type IT not supported!', str(context.exception))

    def test_compare_features_to_reference_error(self):
        # arrange
        features_df = pd.DataFrame()
        circuit_type = 'RB'
        system = 'VF'

        # act
        with self.assertRaises(KeyError) as context:
            Reference.compare_features_to_reference(features_df, circuit_type, system)

        # assert
        self.assertEqual('\'System VF is not supported!\'', str(context.exception))

    def test_compare_difference_of_features_to_reference_error(self):
        # arrange
        features_df = pd.DataFrame()
        reference_df = pd.DataFrame()
        circuit_type = 'RB'
        system = 'EE'
        wildcard = ''

        # act
        with self.assertRaises(KeyError) as context:
            Reference.compare_difference_of_features_to_reference(features_df, reference_df, circuit_type,
                                                                  system, wildcard)

        # assert
        self.assertEqual('\'System EE is not supported!\'', str(context.exception))

    def test_get_mp3_period(self):
        # arrange
        date_time = '2018-01-01 00:00:00.000+01:00'

        # act
        period_act = Reference.get_mp3_period(date_time)

        # assert
        period_exp = 'eYETS 2017/2018'
        self.assertEqual(period_exp, period_act)

    # Nominal voltage-dependent QH discharge reference criteria - 900 V
    def test_read_quench_heaters_reference_features_rb(self):
        # arrange
        circuit_type = 'RB'

        # act
        features_act = Reference.read_quench_heaters_reference_features(circuit_type)

        # assert
        features_ref = pd.DataFrame({'min': {'%CELL%:U_HDS_1:first': 780, '%CELL%:U_HDS_2:first': 780,
                                             '%CELL%:U_HDS_3:first': 780, '%CELL%:U_HDS_4:first': 780,
                                             '%CELL%:U_HDS_1:last20mean': 15, '%CELL%:U_HDS_2:last20mean': 15,
                                             '%CELL%:U_HDS_3:last20mean': 15, '%CELL%:U_HDS_4:last20mean': 15},
                                     'max': {'%CELL%:U_HDS_1:first': 980, '%CELL%:U_HDS_2:first': 980,
                                             '%CELL%:U_HDS_3:first': 980, '%CELL%:U_HDS_4:first': 980,
                                             '%CELL%:U_HDS_1:last20mean': 70, '%CELL%:U_HDS_2:last20mean': 70,
                                             '%CELL%:U_HDS_3:last20mean': 70, '%CELL%:U_HDS_4:last20mean': 70}})
        pd.testing.assert_frame_equal(features_ref, features_act)

    # Support to several reference QH discharges to account for magnet replacement over time
    def test_get_quench_heater_reference_discharge_b24l5(self):
        """This test checks the retrieval of a QH reference for no input timestamp"""
        # arrange
        circuit_type = 'RB'
        magnet_name = 'B24L5'

        # act
        timestamp_act = Reference.get_quench_heater_reference_discharge(circuit_type,
                                                                        magnet_name,
                                                                        1424018504015000000)

        # assert
        timestamp_ref = 1424018504014000000
        self.assertEqual(timestamp_ref, timestamp_act)

    def test_get_quench_heater_reference_discharge_b24l5_equal_to_first_reference(self):
        """This test checks the retrieval of a QH reference for no input timestamp"""
        # arrange
        circuit_type = 'RB'
        magnet_name = 'B24L5'
        timestamp = 1424018504014000000

        # act
        timestamp_act = Reference.get_quench_heater_reference_discharge(circuit_type, magnet_name, timestamp)

        # assert
        timestamp_ref = 1424018504014000000
        self.assertEqual(timestamp_ref, timestamp_act)

    def test_get_quench_heater_reference_discharge_b24l5_before_reference(self):
        """This test checks the retrieval of a QH reference for input timestamp before the first reference.
        A warning should be displayed and the first timestamp returned"""
        # arrange
        circuit_type = 'RB'
        magnet_name = 'B24L5'
        timestamp = 1424018504014000000 - 10

        # act
        # assert
        timestamp_ref = 1424018504014000000
        with warnings.catch_warnings(record=True) as w:
            timestamp_act = Reference.get_quench_heater_reference_discharge(circuit_type, magnet_name, timestamp)
            self.assertEqual('The given is older than timestamp reference QH discharge!', str(w[0].message))
            self.assertEqual(timestamp_ref, timestamp_act)

    def test_get_quench_heater_reference_discharge_b24l5_after_first_reference(self):
        """This test checks the retrieval of a QH reference for input timestamp after the first reference.
        The first timestamp returned"""
        # arrange
        circuit_type = 'RB'
        magnet_name = 'B24L5'
        timestamp = 1424018504014000000 + 10

        # act
        timestamp_act = Reference.get_quench_heater_reference_discharge(circuit_type, magnet_name, timestamp)

        # assert
        timestamp_ref = 1424018504014000000
        self.assertEqual(timestamp_ref, timestamp_act)

    def test_get_quench_heater_reference_discharge_b24l5_equal_to_last_reference(self):
        """This test checks the retrieval of a QH reference for input timestamp after the first reference.
        The first timestamp returned"""
        # arrange
        circuit_type = 'RB'
        magnet_name = 'B24L5'
        timestamp = 1611330989888000000

        # act
        timestamp_act = Reference.get_quench_heater_reference_discharge(circuit_type, magnet_name, timestamp)

        # assert
        timestamp_ref = 1611330989888000000
        self.assertEqual(timestamp_ref, timestamp_act)

    def test_get_quench_heater_reference_discharge_b24l5_after_last_reference(self):
        """This test checks the retrieval of a QH reference for input timestamp after the first reference.
        The first timestamp returned"""
        # arrange
        circuit_type = 'RB'
        magnet_name = 'B24L5'
        timestamp = 1611330989888000000 + 10

        # act
        timestamp_act = Reference.get_quench_heater_reference_discharge(circuit_type, magnet_name, timestamp)

        # assert
        timestamp_ref = 1611330989888000000
        self.assertEqual(timestamp_ref, timestamp_act)

    def test_get_quench_heater_reference_discharge_with_duplicate_reference(self):
        """"""
        # arrange
        circuit_type = 'RB'
        magnet_name = 'C31R2'
        timestamp = 1618509304344000000 - 10

        # act
        timestamp_act = Reference.get_quench_heater_reference_discharge(circuit_type, magnet_name, timestamp)

        # assert
        timestamp_ref = 1422019689013000000
        self.assertEqual(timestamp_ref, timestamp_act)

    def test_get_rq_circuit_parameters(self):
        # arrange
        # act
        rq_circuit_parameters = Reference.get_rq_circuit_parameters()

        # assert
        rq_circuit_parameters_ref = pd.DataFrame({'Circuit name': {0: 'RQD/F.A12', 1: 'RQD/F.A23', 2: 'RQD/F.A34',
                                                                   3: 'RQD/F.A45', 4: 'RQD/F.A56', 5: 'RQD/F.A67',
                                                                   6: 'RQD/F.A78', 7: 'RQD/F.A81'},
                                                  'Number of apertures': {0: 47, 1: 51, 2: 51, 3: 47,
                                                                          4: 47, 5: 51, 6: 51, 7: 47},
                                                  'Beam 1': {0: 'Ext', 1: 'Int', 2: 'Int', 3: 'Int', 4: 'Ext', 5: 'Ext',
                                                             6: 'Ext', 7: 'Int'},
                                                  'Beam 2': {0: 'Int', 1: 'Ext', 2: 'Ext', 3: 'Ext',
                                                             4: 'Int', 5: 'Int', 6: 'Int', 7: 'Ext'},
                                                  'RQF B1/B2': {0: 'Odd/Even', 1: 'Even/Odd', 2: 'Odd/Even',
                                                                3: 'Even/Odd', 4: 'Odd/Even', 5: 'Even/Odd',
                                                                6: 'Odd/Even', 7: 'Even/Odd'},
                                                  'RQD B1/B2': {0: 'Even/Odd', 1: 'Odd/Even', 2: 'Even/Odd',
                                                                3: 'Odd/Even', 4: 'Even/Odd', 5: 'Odd/Even',
                                                                6: 'Even/Odd', 7: 'Odd/Even'},
                                                  'Inductance [H]': {0: 0.2744, 1: 0.2856, 2: 0.2856, 3: 0.2744,
                                                                     4: 0.2744, 5: 0.2856, 6: 0.2856, 7: 0.2744},
                                                  'EE Resistance [ohm]': {0: 0.0066, 1: 0.0077, 2: 0.0077, 3: 0.0066,
                                                                          4: 0.0066, 5: 0.0077, 6: 0.0077, 7: 0.0066}})
        pd.testing.assert_frame_equal(rq_circuit_parameters_ref, rq_circuit_parameters)
